""" add a keycloak authentication class specific to Django Rest Framework """
from typing import Tuple, Dict
import logging

from django.core.exceptions import ObjectDoesNotExist
from django.contrib.auth.models import AnonymousUser
from django.contrib.auth import get_user_model
from rest_framework import (
    authentication,
    exceptions,
)

from .keycloak import keycloak_openid
from .settings import api_settings
from . import __title__

log = logging.getLogger(__title__)
User = get_user_model()


class KeycloakAuthentication(authentication.TokenAuthentication):
    keyword = api_settings.KEYCLOAK_AUTH_HEADER_PREFIX

    def authenticate_credentials(
        self,
        token: str
    ) -> Tuple[AnonymousUser, Dict]:
        """ Attempt to verify JWT from Authorization header with Keycloak """
        log.debug('KeycloakAuthentication.authenticate_credentials')
        try:
            user = None
            # Checks token is active
            decoded_token = keycloak_openid.introspect(token)
            is_active = decoded_token.get('active', False)
            if not is_active:
                raise exceptions.AuthenticationFailed(
                    'invalid or expired token'
                )
            if api_settings.KEYCLOAK_MANAGE_LOCAL_USER is not True:
                log.info(
                    'KeycloakAuthentication.authenticate_credentials: '
                    f'{decoded_token}'
                )
                user = AnonymousUser()
            else:
                username = decoded_token['sub']
                email = decoded_token['email']
                # django stores first_name and last_name as empty strings
                # by default, not None
                first_name = decoded_token.get('given_name', '')
                last_name = decoded_token.get('family_name', '')
                try:
                    user = User.objects.get(username=username)
                    user_values = (user.email, user.first_name, user.last_name,)
                    token_values = (email, first_name, last_name,)
                    if user_values != token_values:
                        user.email = email
                        user.first_name = first_name
                        user.last_name = last_name
                        user.save()
                except ObjectDoesNotExist:
                    log.warn(
                        'KeycloakAuthentication.authenticate_credentials - '
                        f'ObjectDoesNotExist: {username} does not exist'
                    )
                if user is None:
                    user = User.objects.create_user(
                        username,
                        email,
                        None  # if None set_unusable_password() will be called.
                    )
            log.info(
                'KeycloakAuthentication.authenticate_credentials: '
                f'{user} - {decoded_token}'
            )
            return (user, decoded_token)
        except Exception as e:
            log.error(
                'KeycloakAuthentication.authenticate_credentials - '
                f'Exception: {e}'
            )
            raise exceptions.AuthenticationFailed()



import logging
from typing import List

from rest_framework import permissions

from . import __title__

log = logging.getLogger(__title__)

ROLE_USER = 'role:user'
ROLE_SERVICE = 'role:service'
ROLE_ADMIN = 'role:admin'


def _has_required_group(request, required_groups: List[str]) -> bool:
    user_groups = [x.name for x in request.user.groups.all()]
    log.info(
        f'{__name__}._has_required_group - required_groups: '
        f'{required_groups}'
    )
    log.info(
        f'{__name__}._has_required_group - user_groups: '
        f'{user_groups}'
    )
    return bool(
        set(required_groups)
        .intersection(
            set([x.name for x in request.user.groups.all()])
        )
    )


class BaseGroupBasedPermission(permissions.BasePermission):
    required_groups = []

    def has_permission(self, request, view):
        log.info(
            f'{self.__class__.__name__}.has_permission: '
            f'{self.required_groups}'
        )
        return bool(
            request.user and _has_required_group(request, self.required_groups)
        )

    def has_object_permission(self, request, view, obj):
        log.info(
            f'{self.__class__.__name__}.has_object_permission: '
            f'{self.required_groups}'
        )
        return bool(
            request.user and _has_required_group(request, self.required_groups)
        )


class HasAdminGroup(BaseGroupBasedPermission):
    required_groups = [ROLE_ADMIN]


class HasUserGroup(BaseGroupBasedPermission):
    required_groups = [ROLE_USER]


class HasServiceGroup(BaseGroupBasedPermission):
    required_groups = [ROLE_SERVICE]


class HasOwnerGroup(permissions.BasePermission):
    """ validate auth user is obj.owner """
    def has_object_permission(self, request, view, obj):
        required_groups = [str(obj.owner.pk)]
        log.info(
            f'{self.__class__.__name__}.has_object_permission: '
            f'{required_groups}'
        )
        return bool(
            request.user and _has_required_group(request, required_groups)
        )


def _has_required_role(request, required_roles: List[str]) -> bool:
    log.info(
        f'{__name__}._has_required_role - required_roles: '
        f'{required_roles}'
    )
    log.info(
        f'{__name__}._has_required_role - request.roles: '
        f'{request.roles}'
    )
    return bool(set(required_roles).intersection(set(request.roles)))


class BaseRoleBasedPermission(permissions.BasePermission):
    required_roles = []

    def has_permission(self, request, view):
        log.info(
            f'{self.__class__.__name__}.has_permission: '
            f'{self.required_roles}'
        )
        return bool(
            request.user and _has_required_role(request, self.required_roles)
        )

    def has_object_permission(self, request, view, obj):
        log.info(
            f'{self.__class__.__name__}.has_object_permission: '
            f'{self.required_roles}'
        )
        return bool(
            request.user and _has_required_role(request, self.required_roles)
        )


class HasAdminRole(BaseRoleBasedPermission):
    required_roles = [ROLE_ADMIN]


class HasUserRole(BaseRoleBasedPermission):
    required_roles = [ROLE_USER]


class HasServiceRole(BaseRoleBasedPermission):
    required_roles = [ROLE_SERVICE]


class HasOwnerRole(permissions.BasePermission):
    """ validate auth user is obj.owner """
    def has_object_permission(self, request, view, obj):
        required_roles = [str(obj.owner.pk)]
        log.info(
            f'{self.__class__.__name__}.has_object_permission: '
            f'{required_roles}'
        )
        return bool(
            request.user and _has_required_role(request, required_roles)
        )



""" Settings config for the drf_keycloak_auth application """
import datetime
import os

from django.conf import settings
from rest_framework.settings import APISettings

USER_SETTINGS = getattr(settings, 'DRF_KEYCLOAK_AUTH', None)

DEFAULTS = {
    'KEYCLOAK_SERVER_URL': os.getenv('KEYCLOAK_SERVER_URL'),
    'KEYCLOAK_REALM': os.getenv('KEYCLOAK_REALM'),
    'KEYCLOAK_CLIENT_ID': os.getenv('KEYCLOAK_CLIENT_ID'),
    'KEYCLOAK_CLIENT_SECRET_KEY': os.getenv('KEYCLOAK_CLIENT_SECRET_KEY'),
    'KEYCLOAK_AUTH_HEADER_PREFIX':
        os.getenv('KEYCLOAK_AUTH_HEADER_PREFIX', 'Bearer'),
    'KEYCLOAK_MANAGE_LOCAL_USER':
        os.getenv('KEYCLOAK_MANAGE_LOCAL_USER', True),
    'KEYCLOAK_MANAGE_LOCAL_GROUPS':
        os.getenv('KEYCLOAK_MANAGE_LOCAL_GROUPS', False),
}

# List of settings that may be in string import notation.
IMPORT_STRINGS = (
)

api_settings = APISettings(USER_SETTINGS, DEFAULTS, IMPORT_STRINGS)



import logging
from typing import List

from django.utils.functional import SimpleLazyObject
from django.contrib.auth.models import Group

from .keycloak import get_resource_roles
from . import __title__
from .settings import api_settings

log = logging.getLogger(__title__)


class KeycloakMiddleware:

    def __init__(self, get_response):     
        # Django response
        self.get_response = get_response

    def __call__(self, request):
        log.debug('KeycloakMiddleware.__call__')
        return self.get_response(request)

    def process_view(self, request, view_func, view_args, view_kwargs) -> None:
        """
        middleware lifecyle method where view has been determined
        
        SimpleLazyObject required as KeycloakMiddleware is called BEFORE
        KeycloakAuthentication
        """
        log.debug('KeycloakMiddleware.process_view')
        # SimpleLazyObject required as KeycloakMiddleware is called BEFORE KeycloakAuthentication
        request.roles = SimpleLazyObject(lambda: self._bind_roles_to_request(request))

    def _bind_roles_to_request(self, request) -> List[str]:
        """ try to add roles from authenticated keycloak user """
        roles = []
        try:
            roles += get_resource_roles(request.auth)
            roles.append(str(request.user.pk))
        except Exception as e:
            log.warn(
                f'KeycloakMiddleware._bind_roles_to_request - Exception: {e}'
            )

        if api_settings.KEYCLOAK_MANAGE_LOCAL_GROUPS is True:
            groups = self._get_or_create_groups(roles)
            self._refresh_user_groups(request, groups)

        log.info(f'KeycloakMiddleware.bind_roles_to_request: {roles}')
        return roles

    def _get_or_create_groups(self, roles: List[str]) -> List[Group]:
        groups = []
        for role in roles:
            group, created = Group.objects.get_or_create(name=role)
            if created:
                log.info(
                    'KeycloakMiddleware._get_or_create_groups - created: '
                    f'{group.name}'
                )
            else:
                log.info(
                    'KeycloakMiddleware._get_or_create_groups - exists: '
                    f'{group.name}'
                )
            groups.append(group)
        return groups

    def _refresh_user_groups(self, request, groups: List[Group]) -> None:
        try:
            # request.user.groups.clear()
            log.info(
                'KeycloakMiddleware._refresh_user_groups: '
                f'{[x.name for x in groups]}'
            )
            request.user.groups.set(groups)  # clears no longer existing groups
        except Exception as e:
            log.warn(
                f'KeycloakMiddleware._refresh_user_groups - exception: {e}'
            )
